==> gcma-0.1.5-py3-none-any.whl: gcma/wrapper.py <==
import logging
import pickle
from typing import Callable, List, Tuple, Union

import cma
import numpy as np
import treefiles as tf
from cma.evolution_strategy import CMAEvolutionStrategyResult, CMAEvolutionStrategy

from gcma.io.BaseIO import IOEncoder
from gcma.io.Feature import TFeatures, Features
from gcma.io.History import History
from gcma.io.Options import Options
from gcma.io.Param import TParams, Params
from gcma.io.utils import load_legends


class CMA:
    """
    Wraps the pycma implementation of N.Hansen: https://github.com/CMA-ES/pycma
    """

    def __init__(
        self,
        params: TParams,
        features: TFeatures,
        options: Options,
        output_dir: [tf.Tree, str],
        energy: Callable = None,
        energy_sum: Callable = None,  # sums up features error
        sigma0: float = 2,
        scale_factor: float = 10,
        plot: bool = True,
        result_type: str = "xbest",
        **kwargs,
    ):
        self.output_dir = self._init_output_dir(output_dir)

        self.params = Params(params)
        self.features = Features(features)
        self.dump_legends()

        self.sigma0 = sigma0
        self.scale_factor = scale_factor
        self.plot = plot
        self.result_type = result_type
        self._finished = False

        self.param_history = []
        self.energy_history = []
        self.energy_history_ft = []
        self.features_history = []
        self.history = History()

        if energy is not None:
            self.energy = energy

        if energy_sum is not None:
            self.energy_sum = energy_sum

        self.es = None
        options.update(kwargs)
        self._init(**options)

        self.plotter = None
        self._init_plotter()

    @classmethod
    def from_previous(cls, path: Union[str, tf.Tree]):
        out = CMA._init_output_dir(path)
        params, features = load_legends(out.legends)
        options = tf.load_json(out.options)

        c = cls(params, features, options, path)
        with open(out.pickle, "rb") as f:
            c.es = pickle.loads(f.read())
        c.es.stop().clear()

        c.param_history = np.load(out.pa).tolist()
        c.energy_history = np.load(out.en).tolist()
        c.energy_history_ft = np.load(out.enf).tolist()
        c.features_history = np.load(out.fv).tolist()

        c.history = History.load(out.history)

        return c

    def _init_plotter(self):
        from gcma.plotter.Plotter import Plotter

        self.plotter = Plotter(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._finished:
            self._finish()

    @property
    def n_params(self):
        return len(self.params)

    @property
    def n_features(self):
        return len(self.features)

    @property
    def popsize(self):
        return self.es.opts["popsize"]

    @property
    def n_iters(self):
        return self.es.opts["maxiter"]

    @staticmethod
    def _init_output_dir(tree) -> tf.Tree:
        tree = tf.Tree(tree)
        tree.file(
            pa="cma_params_history.npy",
            en="cma_energy_history.npy",
            enf="cma_energy_history_ft.npy",
            fv="cma_features_history.npy",
            pickle="cma_pickle.pkl",
            legends="legends.json",
            options="options.json",
            history="history.json",
        )
        return tree

    @property
    def bounds(self) -> np.ndarray:
        return np.array([x.bounds for x in self.params.values()])

    def _backup(self):
        best_i = np.argmin(self.energy_history[-1])

        # Best set of parameters
        # equivalent to `self.es.best.last.x`
        bestx = self.param_history[-1][best_i, ...]
        params = self.params.assign(self.rescale(bestx))
        assert np.array_equal(bestx, self.es.best.last.x)

        # Features values associated to best parameters
        bestfv = self.features_history[-1][best_i, ...]
        features = self.features.assign(bestfv)

        # Features errors associated to best parameters
        bestfe = self.energy_history_ft[-1][best_i, ...]
        features = features.assign(bestfe, "error")

        # Global energy associated to best parameters
        # equivalent to `self.es.best.last.f`
        bestf = self.energy_history[-1][best_i]
        assert np.array_equal(bestf, self.es.best.last.f)

        self.history.add(params=params, features=features, fvalue=bestf)

    def get_history(self, t: str):
        if len(self.history) == 0:
            return "No history to display"
        return getattr(self.history, t)

    @property
    def last(self):
        return self.get_history("last")

    @property
    def best(self):
        return self.get_history("best")

    def _init(self, **options):
        initial_values = np.array([x.initial_value for x in self.params.values()])
        n_val = len(self.bounds)

        if initial_values is None:
            initial_values = [self.scale_factor / 2] * n_val
        else:
            initial_values = self.descale(initial_values)

        inopts = {
            "bounds": ([0] * n_val, [self.scale_factor] * n_val),
            "verb_filenameprefix": f"{self.output_dir.abs()}/",
            "termination_callback": self.termination_callback,
        }
        options.update(inopts)

        self.es = cma.CMAEvolutionStrategy(
            x0=initial_values, sigma0=self.sigma0, inopts=options
        )
        self.dump_options()

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def termination_callback(self, es: CMAEvolutionStrategy) -> bool:
        return False

    def dump_legends(self):
        tf.dump_json(
            self.output_dir.legends,
            {"params": self.params, "features": self.features},
            cls=IOEncoder,
        )

    def load_legends(self) -> Tuple[TParams, TFeatures]:
        return load_legends(self.output_dir.legends)

    def dump_options(self):
        options = {k: self.es.opts[k] for k in Options.to_save()}
        tf.dump_json(self.output_dir.options, options)

    def rescale(self, x):
        """
        From normalized CMA-ES values (in [0, scale_factor]) to real world
        """
        b, s = self.bounds, self.scale_factor
        return x * (b[:, 1] - b[:, 0]) / s + b[:, 0]

    def descale(self, x):
        """
        Normalize from real world coords to CMAES world (btw 0 and scale_factor)
        """
        b, s = self.bounds, self.scale_factor
        return (x - b[:, 0]) * s / (b[:, 1] - b[:, 0])

    def energy(self, params: List[TParams], **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """
        Main method to compute the errors related to each sample features
        Return a tuple of 2 arrays of size (popsize, n_features), the
        first one with the features error, the second with features values
        """
        raise NotImplementedError

    def energy_sum(self, features_error: np.ndarray) -> np.ndarray:
        """
        Main method to concatenate the features errors for each sample
        Return an array of size (popsize,)
        """
        assert features_error.shape == (self.popsize, self.n_features)
        return features_error.sum(axis=1)

    def _generation(self):
        values_to_test = np.array(self.es.ask())
        self.param_history.append(values_to_test)
        scaled = self.rescale(values_to_test)

        params = []  # list of size `self.popsize`. Each sublist contains the parameters
        for i in range(self.popsize):
            new_set = self.params.copy()
            for j, p in enumerate(new_set.values()):
                p.value = scaled[i, j]
            params.append(new_set)

        # log.info(f"CMAES iteration={es.countiter}")

        features_error, features_value = self.energy(params, n_iter=self.es.countiter)
        self.energy_history_ft.append(features_error)
        self.features_history.append(features_value)

        # features = []  # list of size `self.popsize`. Each sublist contains the features
        # for i in range(self.popsize):
        #     new_set = self.features.copy()
        #     for j, p in enumerate(new_set):
        #         p.value = results[i, j]
        #     features.append(new_set)

        samples_error = self.energy_sum(features_error)
        self.energy_history.append(samples_error)

        self.es.tell(values_to_test, samples_error)
        self.es.logger.add()
        self.es.disp()

        self._backup()

    def run(self, *exceptions) -> CMAEvolutionStrategyResult:
        self._finished = False

        while not self.es.stop():
            if len(exceptions) == 0:
                self._generation()
            else:
                try:
                    self._generation()
                except exceptions as e:
                    log.critical(f"Exiting CMA-ES main loop: {e}")
                    break

        self._finish()

        return self.es.result

    def _finish(self):
        with open(self.output_dir.pickle, "wb") as f:
            f.write(self.es.pickle_dumps())

        np.save(self.output_dir.pa, self.param_history)
        np.save(self.output_dir.en, self.energy_history)
        np.save(self.output_dir.enf, self.energy_history_ft)
        np.save(self.output_dir.fv, self.features_history)

        self.history.dump(self.output_dir.history)

        log.info(f"Finishing CMA-ES ({self.es.countiter} iterations)")
        self._finished = True

    def set_opts(self, **kwargs):
        self.es.opts.update(kwargs)
        self.dump_options()

    @property
    def options(self):
        return self.es.opts

    @options.setter
    def options(self, d: dict):
        self.set_opts(**d)


log = logging.getLogger(__name__)

==> gcma-0.1.18-py3-none-any.whl: gcma/wrapper.py <==
import logging
import pickle
from datetime import datetime
from typing import List, Tuple, Union

import cma
import numpy as np
import treefiles as tf
from cma.evolution_strategy import CMAEvolutionStrategyResult, CMAEvolutionStrategy
from gcma.constraints import Constraints, TConstraints
from gcma.io.BaseIO import IOEncoder
from gcma.io.Feature import TFeatures, Features
from gcma.io.History import History
from gcma.io.Options import Options
from gcma.io.Param import TParams, Params
from gcma.io.Termination import TermOptions
from gcma.io.utils import load_legends


class CMA:
    """
    Wraps the pycma implementation of N.Hansen: https://github.com/CMA-ES/pycma
    """

    def __init__(
        self,
        params: TParams,
        features: TFeatures,
        options: Options,
        output_dir: [tf.Tree, str],
        sigma0: float = 2,
        scale_factor: float = 10,
        plot: bool = True,
        result_type: str = "xbest",
        termination_options: TermOptions = None,
        constraints: TConstraints = None,
        **kwargs,
    ):
        self.output_dir = self._init_output_dir(output_dir)

        self.params = Params(params)
        self.features = Features(features)
        self.dump_legends()

        self.sigma0 = sigma0
        self.scale_factor = scale_factor
        self.plot = plot
        self.result_type = result_type
        self.termination_options = (
            TermOptions() if termination_options is None else termination_options
        )
        self.constraints = Constraints(constraints)
        self._finished = False

        # fmt: off
        self.param_history = []  # parameters values associated to each sample
        self.features_history = []  # values of the features associated to the best sample
        self.energy_history_ft = []  # errors of the features associated to sample
        self.energy_history = []  # global energy associated to each sample (score + constraint + prior)
        self.energy_history_score = []  # score value associated to each sample
        self.energy_history_constraints = []  # constraint value associated to each sample
        self.energy_history_prior = []  # regularization (prior) value associated to each sample
        # fmt: on
        self.history = History()

        self.es = None
        options.update(kwargs)
        self._init(**options)

        self.plotter = None
        self._init_plotter()

    @classmethod
    def from_previous(cls, path: Union[str, tf.Tree]):
        out = CMA._init_output_dir(path)
        params, features = load_legends(out.legends)
        options = tf.load_json(out.options)
        c = cls(params, features, options, path)
        c.init_previous()
        return c

    def init_previous(self):
        with open(self.output_dir.pickle, "rb") as f:
            self.es = pickle.loads(f.read())
        self.es.stop().clear()

        self.param_history = np.load(self.output_dir.pa).tolist()
        self.energy_history = np.load(self.output_dir.en).tolist()
        self.energy_history_ft = np.load(self.output_dir.enf).tolist()
        self.energy_history_score = np.load(self.output_dir.score).tolist()
        self.energy_history_constraints = np.load(self.output_dir.cts).tolist()
        self.energy_history_prior = np.load(self.output_dir.prior).tolist()
        self.features_history = np.load(self.output_dir.fv).tolist()

        self.termination_options = TermOptions.from_dict(
            tf.load_json(self.output_dir.term)
        )
        self.termination_options["term_on_date"] = False

        self.history = History.load(self.output_dir.history)

    def _init_plotter(self):
        from gcma.plotter.Plotter import Plotter

        self.plotter = Plotter(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._finished:
            self._finish()

    @property
    def n_params(self):
        return len(self.params)

    @property
    def n_features(self):
        return len(self.features)

    @property
    def popsize(self):
        return self.es.opts["popsize"]

    @property
    def n_iters(self):
        return self.es.opts["maxiter"]

    @staticmethod
    def _init_output_dir(tree) -> tf.Tree:
        tree = tf.Tree(tree)
        tree.file(
            pa="cma_params_history.npy",
            en="cma_energy_history.npy",
            enf="cma_energy_history_ft.npy",
            score="cma_energy_history_score.npy",
            cts="cma_energy_history_constraints.npy",
            prior="cma_energy_history_prior.npy",
            fv="cma_features_history.npy",
            pickle="cma_pickle.pkl",
            legends="legends.json",
            options="options.json",
            history="history.json",
            term="termination_options.json",
        )
        return tree

    @property
    def bounds(self) -> np.ndarray:
        return np.array([x.bounds for x in self.params.values()])

    def _backup(self):
        best_i = np.argmin(self.energy_history[-1])

        # Best set of parameters
        # equivalent to `self.es.best.last.x`
        bestx = self.param_history[-1][best_i, ...]
        params = self.params.assign(self.rescale(bestx))
        assert np.array_equal(bestx, self.es.best.last.x)

        # Features values associated to best parameters
        bestfv = self.features_history[-1][best_i, ...]
        features = self.features.assign(bestfv)

        # Features errors associated to best parameters
        bestfe = self.energy_history_ft[-1][best_i, ...]
        features = features.assign(bestfe, "error")

        # Global energy associated to best parameters
        # equivalent to `self.es.best.last.f`
        bestf = self.energy_history[-1][best_i]
        assert np.array_equal(bestf, self.es.best.last.f)

        self.history.add(params=params, features=features, fvalue=bestf)

    def get_history(self, t: str):
        if len(self.history) == 0:
            return "No history to display"
        return getattr(self.history, t)

    @property
    def last(self):
        return self.get_history("last")

    @property
    def best(self):
        return self.get_history("best")

    def _init(self, **options):
        initial_values = np.array([x.initial_value for x in self.params.values()])
        n_val = len(self.bounds)

        if initial_values is None:
            initial_values = [self.scale_factor / 2] * n_val
        else:
            initial_values = self.descale(initial_values)

        inopts = {
            "bounds": ([0] * n_val, [self.scale_factor] * n_val),
            "verb_filenameprefix": f"{self.output_dir.abs()}/",
            "termination_callback": self.termination_callback,
        }
        options.update(inopts)

        self.es = cma.CMAEvolutionStrategy(
            x0=initial_values, sigma0=self.sigma0, inopts=options
        )
        self.dump_options()

    def termination_callback(self, es: CMAEvolutionStrategy) -> bool:
        opts = self.termination_options

        is_over = opts.is_over()
        is_tol = False
        if es.countiter > 1:
            values = [x.error for x in self.history.best.features.values()]
            is_tol = opts.is_tol(values)

        return is_over or is_tol

    def dump_legends(self):
        tf.dump_json(
            self.output_dir.legends,
            {"params": self.params, "features": self.features},
            cls=IOEncoder,
        )

    def load_legends(self) -> Tuple[TParams, TFeatures]:
        return load_legends(self.output_dir.legends)

    def dump_options(self):
        options = {k: self.es.opts[k] for k in Options.to_save()}
        tf.dump_json(self.output_dir.options, options)

    def rescale(self, x):
        """
        From normalized CMA-ES values (in [0, scale_factor]) to real world
        """
        b, s = self.bounds, self.scale_factor
        return x * (b[:, 1] - b[:, 0]) / s + b[:, 0]

    def descale(self, x):
        """
        Normalize from real world coords to CMAES world (btw 0 and scale_factor)
        """
        b, s = self.bounds, self.scale_factor
        return (x - b[:, 0]) * s / (b[:, 1] - b[:, 0])

    def energy(self, params: List[TParams], **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """
        Main method to compute the errors related to each sample features
        Return a tuple of 2 arrays of size (popsize, n_features), the
        first one with the features error, the second with features values
        """
        raise NotImplementedError

    def energy_sum(self, features_error: np.ndarray) -> np.ndarray:
        """
        Main method to concatenate the features errors for each sample
        Return an array of size (popsize,)
        """
        assert features_error.shape == (self.popsize, self.n_features)
        return features_error.sum(axis=1)

    def energy_constraints(self, params: List[TParams]) -> np.ndarray:
        """
        Main method to concatenate the constraints contributions for each sample
        Return an array of size (popsize,), each row contains the sum of all constraints
        """
        cts = np.zeros(self.popsize)

        for i, p in enumerate(params):
            contribs = self.constraints.apply(p)
            contribs = np.array([x[1] for x in contribs])
            cts[i] = contribs.sum()

        return cts

    def _generation(self):
        # log.info(f"CMAES iteration={self.es.countiter}")

        values_to_test = np.array(self.es.ask())
        self.param_history.append(values_to_test)
        scaled = self.rescale(values_to_test)

        # list of size `self.popsize`. Each sublist contains the parameters
        params = [self.params.assign(scaled[i, :]) for i in range(self.popsize)]

        features_error, features_value = self.energy(params, n_iter=self.es.countiter)
        self.energy_history_ft.append(features_error)
        self.features_history.append(features_value)

        samples_error = self.energy_sum(features_error)
        self.energy_history_score.append(samples_error)

        contraints = self.energy_constraints(params)
        self.energy_history_constraints.append(contraints)

        prior_regul = self.energy_prior()
        self.energy_history_prior.append(prior_regul)

        energy = samples_error + contraints + prior_regul
        self.energy_history.append(energy)

        self.es.tell(values_to_test, energy)
        self.es.logger.add()
        self.es.disp()

        self._backup()
        self.end_generation()

    def energy_prior(self) -> np.ndarray:
        # TODO
        return np.zeros(self.popsize)

    def run(self, *exceptions) -> CMAEvolutionStrategyResult:
        self._finished = False

        while not self.es.stop():
            if len(exceptions) == 0:
                self._generation()
            else:
                try:
                    self._generation()
                except exceptions as e:
                    log.critical(f"Exiting CMA-ES main loop: {e}")
                    break

        self._finish()

        return self.es.result

    def end_generation(self):
        """
        Called at the end of each CMA-ES iteration for cleaning purposes
        """
        pass

    def _finish(self):
        with open(self.output_dir.pickle, "wb") as f:
            f.write(self.es.pickle_dumps())

        np.save(self.output_dir.pa, self.param_history)
        np.save(self.output_dir.en, self.energy_history)
        np.save(self.output_dir.enf, self.energy_history_ft)
        np.save(self.output_dir.score, self.energy_history_score)
        np.save(self.output_dir.cts, self.energy_history_constraints)
        np.save(self.output_dir.prior, self.energy_history_prior)
        np.save(self.output_dir.fv, self.features_history)

        self.history.dump(self.output_dir.history)
        tf.dump_json(self.output_dir.term, self.termination_options)

        with open(self.output_dir.path("best_results.txt"), "w") as f:
            f.write(f"Time: {datetime.now()}\n")
            f.write(f"Stop criteria: {self.es.stop()}\n")
            f.write(f"Total iterations: {self.es.countiter}\n")
            f.write(f"Best Result:\n{self.best}")

        log.info(f"Finishing CMA-ES ({self.es.countiter} iterations)")
        self._finished = True

    def set_opts(self, **kwargs):
        self.es.opts.update(kwargs)
        self.dump_options()

    @property
    def options(self):
        return self.es.opts

    @options.setter
    def options(self, d: dict):
        self.set_opts(**d)

    def zip(self):
        fname = self.output_dir.path(f"{self.output_dir.basename}.zip")
        tf.dump_zip(fname, self.output_dir)
        log.info(f"Saved zip to {fname!r}")


log = logging.getLogger(__name__)
